unset_jrh_lexer;;
let unix_pfinet = Unix.PF_INET;;
let unix_sockstream = Unix.SOCK_STREAM;;
let unix_addrinet (a, b) = Unix.ADDR_INET (a,b);;
let unix_wnohang = Unix.WNOHANG;;
let unix_uninet (Unix.ADDR_INET (a,b)) = (a,b);;
let unix_f_ulock = Unix.F_ULOCK;;
let unix_f_lock  = Unix.F_LOCK;;
let unix_f_rlock = Unix.F_RLOCK;;
set_jrh_lexer;;

type ('a, 'b) sum = Inl of 'a | Inr of 'b | Err of int;;

let subprocesses = ref [];;

let handle_term _ =
  List.iter (fun i -> try Unix.kill i Sys.sigterm with _ -> ()) !subprocesses; exit 0;;
Sys.set_signal Sys.sigterm (Sys.Signal_handle handle_term);;

let run_parallel (progress_fn : 'a -> unit) (sec_fn : unit -> unit)
    time (lst : (('a -> unit) -> 'b) list) =
  let piper, pipew = Unix.pipe () in
  let start f =
    let pid = Unix.fork () in
    let oc = Unix.out_channel_of_descr pipew in
    if pid = 0 then begin try
      Unix.close piper;
      let progress_sub_fn a =
        output_value oc (Inl a); flush oc
      in
      let ret = f progress_sub_fn in
      output_value oc (Inr ret);
      flush oc;
      exit 0
    with _ ->
      try output_value oc (Err (Unix.getpid ())); flush oc; exit 0
      with _ -> exit 0
    end; pid
  in
  subprocesses := List.map start lst;
  Unix.close pipew;
  let rec select desc time =
    if time <= 0. then 0. else
    let (r, _, _) = Unix.select [desc] [] [] 1. in
    if r <> [] then time else (sec_fn (); select desc (time -. 1.))
  in
  Unix.set_nonblock piper;
  let inc = Unix.in_channel_of_descr piper in
  let rec ret time =
    if !subprocesses = [] then None else
    let interp time = function
      | Inl pr -> progress_fn pr; ret time
      | Inr ret -> Some (ret)
      | Err pid ->
          subprocesses := List.filter (fun i -> i <> pid) !subprocesses;
          ignore (Unix.waitpid [] pid);
          ret time
    in
    try interp time (input_value inc) with Sys_blocked_io | Unix.Unix_error _ ->
    let ntime = select piper time in
    if ntime > 0. then interp ntime (input_value inc) else None
  in
  let ret = ret time in
  List.iter (fun i -> try Unix.kill i Sys.sigterm with _ -> ()) !subprocesses;
  Unix.close piper;
  List.iter (fun i -> try ignore (Unix.waitpid [] i) with _ -> ()) !subprocesses;
  subprocesses := [];
  (ret : 'b option)
;;

let knn kno symn max atp syms =
  let fnamei = Filename.temp_file ~temp_dir:"/home/ami/cek/www/hhdebug/" "knni" ".txt" in
  let oc = Pervasives.open_out fnamei in
  output_char oc '\"'; oiter oc output_string "\", \"" syms; output_string oc "\"\n"; close_out oc;
  let fnameo = Filename.temp_file ~temp_dir:"/home/ami/cek/www/hhdebug/" "knno" ".txt" in
  let symext = if symn = 0 then "0" else if symn = 1 then "n" else "t" in
  let aext = if atp then ".atp" else ".human" in
  let dir = Sys.getcwd () ^ "/HH" in
  ignore (Sys.command (Printf.sprintf
    "/home/ami/cek/bin/knn %i \"%s/syms%s\" \"%s/deps%s\" %i \"%s/seq\" < %s > %s 2>/dev/null" kno dir symext dir aext max dir fnamei fnameo));
  let ic = Pervasives.open_in fnameo in
  let adv = string_list_of_string (input_line ic) ' ' in
  close_in ic;
  adv
;;
let nb symn max atp syms =
  let fnamei = Filename.temp_file ~temp_dir:"/home/ami/cek/www/hhdebug/" "nbi" ".txt" in
  let oc = Pervasives.open_out fnamei in
  output_char oc '\"'; oiter oc output_string "\", \"" syms; output_string oc "\"\n"; close_out oc;
  let fnameo = Filename.temp_file ~temp_dir:"/home/ami/cek/www/hhdebug/" "nbo" ".txt" in
  let symext = if symn = 0 then "0" else if symn = 1 then "n" else "t" in
  let aext = if atp then ".atp" else ".human" in
  let dir = Sys.getcwd () ^ "/HH" in
  ignore (Sys.command (Printf.sprintf
    "/home/ami/cek/bin/nbayes %s/syms%s %s/deps%s %i %s/seq < %s > %s 2>/dev/null" dir symext dir aext max dir fnamei fnameo));
  let ic = Pervasives.open_in fnameo in
  let adv = string_list_of_string (input_line ic) ' ' in
  close_in ic;
  adv
;;

let run_prover fname prover time =
  let time = string_of_int time in
  let cmd = match prover with
    1 -> "vampire --mode casc -t " ^ time ^ " --proof tptp --output_axiom_names on " ^ fname
  | 2 -> "z3 -tptp DISPLAY_UNSAT_CORE=true -T:"^time^" " ^ fname
  | 3 -> "eproof --cpu-limit="^time^" --tstp-format " ^ fname
  | _ -> "/home/ami/cek/hhrepo/HH/runepar.pl "^time^" 0 " ^ fname ^ " 2 1"
  in
  let fnameo = fname ^ string_of_int prover ^ "o" in
  let cmd = cmd ^ " > " ^ fnameo ^ " 2> /dev/null" in
  ignore (Sys.command cmd);
  let ret = Sys.command ("grep -q 'status Theorem' " ^ fnameo) in
  if ret <> 0 then failwith "No proof" else
  let cmd = (if prover = 2 then "../../HH/depsz3.sh " else if prover = 1 then "../../HH/depsv.sh " else "../../HH/deps.sh ") ^ fnameo ^ " > " ^ fnameo ^ "d" in
  ignore (Sys.command cmd);
  string_list_of_string (string_of_file (fnameo ^ "d")) ' '
;;

let tryerr f a e = try f a with _ -> failwith e;;

let count_time stime =
  Printf.sprintf "%.2f" (Unix.gettimeofday () -. stime);;

let strats = [
  (("knn-k040h-log", knn  40), (2, 0,  128, false));
  (("knn-k040h-log", knn  40), (2, 0, 1024, false));
  (("knn-k040h-log", knn  40), (2, 2,   32, false));
  (("knn-k760a-log", knn 760), (2, 1,  128,  true));
  (("knn-k040a-log", knn  40), (2, 0,  128,  true));
  (("knn-k040a-log", knn  40), (2, 0, 1024,  true));
  (("knn-k040a-log", knn 760), (2, 0,   96,  true))
];;

let prove_asm_gl ((asm, gl), tac)=
  let gl = itlist (curry mk_imp) asm gl in
  prove(gl, EVERY (replicate DISCH_TAC (List.length asm)) THEN tac)
;;

let hh_svr_dp ((asm, gl) as tm2) =
  let stime = Unix.gettimeofday () in
  let jobs = [(fun _ -> ignore (prove_asm_gl(tm2, CONV_TAC TAUT)); "CONV_TAC TAUT")
   ; (fun _ -> ignore(prove_asm_gl(tm2, INT_ARITH_TAC)); "INT_ARITH_TAC")
   ; (fun _ -> ignore(prove_asm_gl(tm2, SIMP_TAC[ARITH])); "SIMP_TAC[ARITH]")
] in
  let jobs = if asm = [] then jobs else
    (fun _ -> ignore(prove_asm_gl(tm2, ASM_INT_ARITH_TAC)); "ASM_INT_ARITH_TAC") ::
 jobs in
  match run_parallel (fun () -> ()) (fun () -> ()) 30. jobs with
    Some s -> ("<br/>SUCCESS (" ^ count_time stime ^ "s): " ^ s)
  | None -> failwith "hh_decision_procedure failed";;

let find_hash = Hashtbl.find (get_hashes ());;
let find_thm = Hashtbl.find (get_thm_names ());;
let find_hthm n = find_thm (find_hash n);;

let hh_svr_atp time (tm, asm, gl) ((name, advfun), (symn, prover, cut, atp)) pr_fun =
  let (sym0, sym, symt, _, _) = tryerr get_symbols tm "get_symbols failed" in
  pr_fun "S";
  let syms = if symn = 0 then sym0 else if symn = 1 then sym else symt in
  let adv = advfun symn cut atp syms in
  let proverc = function 1 -> 'V' | 2 -> 'Z' | 3 -> 'e' | _ -> 'E' in
  let name = Printf.sprintf "%s-%04i%c" name cut (proverc prover)  in
  let fname = Filename.temp_file ~temp_dir:"/home/ami/cek/www/hhdebug/" ("atpin_" ^ name ^ "_") ".p" in
  let adv1 = rev (List.fold_left (fun l n -> try (find_hthm n, n) :: l with Not_found -> l) [] adv) in
  let (adv1t, adv1) = List.split adv1 in
  let (adv2, _) = List.fold_left (fun (l, no) _ -> (("GOALPREMISE" ^ string_of_int no) :: l, no + 1))
    ([], 0) asm in
  let adv2t = List.map (fun x -> mk_fthm ([], x)) asm in
  pr_fun "A";
  write_atp_proof fname (adv2t @ adv1t, adv2 @ adv1) "conjecture" gl;
  Unix.chmod fname 0o644;
  pr_fun "W";
  run_prover fname prover time
;;

let rec hh_minimize res sec_fn (asm, gl) =
  let fname = "/tmp/atpm" ^ string_of_int (Hashtbl.hash (res, asm, gl)) ^ ".p" in
  let adv = rev (List.fold_left (fun l n -> try (find_hthm n, n) :: l with Not_found ->
    if String.length n > 11 && String.sub n 0 11 = "GOALPREMISE" then
      let i = int_of_string (String.sub n 11 (String.length n - 11)) in
      (mk_fthm ([], (List.nth asm i)), n) :: l else failwith "minimize bug!")
      [] res) in
  write_atp_proof fname (List.split adv) "conjecture" gl;
  let minimize_prover p _ =
    let ret = run_prover fname p 30 in
    let len = List.length ret in
    if len < List.length res then ret else failwith "same"
  in
  match run_parallel (fun () -> ()) (fun () -> ()) 10.
    [minimize_prover 0; minimize_prover 1; minimize_prover 3] with
    | None -> res
    | Some ret -> hh_minimize ret (fun () -> ()) (asm, gl)
;;

let clean_names =
  let regxp = Str.regexp "_conjunct[0-9]*" in
  let clean_name s = Str.global_replace regxp "" s in
  fun l -> uniq (sort (fun x y -> compare x y > 0) (List.map clean_name l));;

let hh_replay ret sec_fn (asm, gl) =
  let pnames, tnames = List.partition
    (fun n -> String.length n > 11 && String.sub n 0 11 = "GOALPREMISE") ret in
  let tnames = clean_names (List.map find_hash tnames) in
  let tnameb = "[" ^ String.concat ";" tnames ^ "]" in
  let ths = List.map find_thm tnames in
  let jobs =
    [(fun _ -> ignore(prove_asm_gl((asm,gl), REWRITE_TAC ths)); ("REWRITE_TAC" ^ tnameb));
     (fun _ -> ignore(prove_asm_gl((asm,gl), SIMP_TAC ths)); ("SIMP_TAC" ^ tnameb));
     (fun _ -> verbose:=false;ignore(prove_asm_gl((asm,gl), MESON_TAC ths)); ("MESON_TAC" ^ tnameb))] in
  let jobs = if pnames = [] then jobs else jobs @
    [(fun _ -> ignore(prove_asm_gl((asm,gl), ASM_REWRITE_TAC ths)); ("ASM_REWRITE_TAC" ^ tnameb));
     (fun _ -> ignore(prove_asm_gl((asm,gl), ASM_SIMP_TAC ths)); ("ASM_SIMP_TAC" ^ tnameb));
     (fun _ -> verbose:=false;ignore(prove_asm_gl((asm,gl), ASM_MESON_TAC ths)); ("ASM_MESON_TAC" ^ tnameb))] in
  match run_parallel (fun () -> ()) sec_fn 10. jobs with
    None -> failwith "Reconstruction failed"
  | Some s -> s
;;

let cache_oc = ref stdout;;
let prog str st =
  Printf.fprintf !cache_oc "<span class='progress'>(%s:%s)</span>%!" str
     (if st then "<span class='ok'>OK</span>" else "?");;
let cachewr s = Printf.fprintf !cache_oc "%s%!" s;;


let hh_svr time (tm, asm, gl) =
  let stime = Unix.gettimeofday () in
  let try_atps prfun =
    let jobs = List.map (hh_svr_atp time (tm, asm, gl)) strats in
    match run_parallel (fun pr -> prfun (false, pr)) (fun () -> ()) (float_of_int time) jobs with
      None -> failwith "No advice"
    | Some lret ->
        let find_hash s = try find_hash s with _ -> s in
        prfun (true, ("<br/>Result (" ^ count_time stime ^ "s): " ^ (String.concat " " (List.map find_hash lret))));
        let mret = hh_minimize lret (fun () -> ()) (asm, gl) in
        if mret <> lret then prfun (true, "<br/>Minimized: " ^ String.concat " " (List.map find_hash mret));
        let stime = Unix.gettimeofday () in
        let repl = hh_replay mret (fun () -> ()) (asm, gl) in
        ("<br/>Replaying: SUCCESS (" ^ count_time stime ^ "s):" ^ repl)
  in
  let try_dps _ = hh_svr_dp (asm, gl) in
  let pfun = function (false, s) -> Printf.fprintf !cache_oc "<span class='progress'>%s</span>%!" s
                     | (true, s) -> Printf.fprintf !cache_oc "%s\n%!" s in
  match run_parallel pfun (fun () -> ()) 120.
    [try_dps; try_atps] with
      Some s -> Printf.fprintf !cache_oc "%s`%!" s
    | None -> failwith "No more advice"
;;

let decode_char = function
  '0' -> 0 | '1' -> 1 | '2' -> 2 | '3' -> 3 | '4' -> 4 | '5' -> 5 | '6' -> 6 | '7' -> 7
| '8' -> 8 | '9' -> 9 | 'A' -> 10| 'B' -> 11| 'C' -> 12| 'D' -> 13| 'E' -> 14| 'F' -> 15
| _ -> failwith "decode_char";;

let decode s =
  let n = ref 0 and l = String.length s in
  for i = 0 to l - 1 do
    if String.unsafe_get s i = '%' then incr n;
  done;
  if !n = 0 then s else
    let s' = String.create (l - 2 * !n) in
    let opos, npos = ref 0, ref 0 in
    let getchar () =
      let c1 = decode_char (String.unsafe_get s (!opos + 1))
      and c2 = decode_char (String.unsafe_get s (!opos + 2)) in
      Char.chr (16 * c1 + c2)
    in
    while !opos < l do
      match String.unsafe_get s !opos with
        '%' -> String.unsafe_set s' !npos (getchar ()); opos := !opos + 3; incr npos
      | x -> String.unsafe_set s' !npos x; incr opos; incr npos
    done;
    s';;

let encode_char = function
  0 -> '0' | 1 -> '1' | 2 -> '2' | 3 -> '3' | 4 -> '4' | 5 -> '5' | 6 -> '6' | 7 -> '7'
| 8 -> '8' | 9 -> '9' |10 -> 'A' | 11 -> 'B'| 12 -> 'C'| 13 -> 'D'| 14 -> 'E'|15 -> 'F'
| _ -> failwith "encode_char";;

let encode s =
  let b = Buffer.create (String.length s * 2) in
  let add_char c =
    let cc = Char.code c in
    if ((cc >= 48 && cc <= 57)
     or (cc >= 65 && cc <= 90)
     or (cc >= 97 && cc <= 122)
     or cc = 45 or cc = 95) then Buffer.add_char b c
    else (Buffer.add_char b '%'; Buffer.add_char b (encode_char (cc / 16));
          Buffer.add_char b (encode_char (cc mod 16))) in
  String.iter add_char s;
  Buffer.contents b;;

let main_cache_dir = !hol_dir ^ "/../../cache/";;
let cache_dir = "HH/cache/";;

let hh_svr_line ln_orig =
  let cache1 = cache_dir ^ Digest.to_hex (Digest.string ln_orig) in
  if Sys.file_exists cache1 && string_of_file cache1 <> "<span class='progress'>(starting)</span>"
    && string_of_file cache1 <> "." then raise Exit;
  (try Unix.unlink cache1 with _ -> ());
  cache_oc := Pervasives.open_out cache1;
  try
    prog "session" true;
    let ln = decode ln_orig in
    let sl = string_list_of_string ln (Char.chr 96) in
    let s2 = String.concat " ==> " sl in
    let (pt, _) = tryerr (parse_preterm o lex o explode) s2 "parse_preterm failed" in
    let pt2 = tryerr (retypecheck []) pt "retypecheck failed" in
    let tm = tryerr term_of_preterm pt2 "term_of_preterm failed" in
    let (asm_r, gl) = tryerr (List.fold_left (fun (ps, c) _ -> let (l, r) = dest_imp c in
      (l :: ps, r)) ([], tm)) (List.tl sl) "split assumptions failed" in
    let asm = List.rev asm_r in
    prog "parse" true;
    let tm_std = rename_all (retyvar tm) prfunnt in
    let md5 = md5 (rename_consts hash_const_name_concat tm_std) in
    close_out !cache_oc;
    (try Unix.unlink cache1 with _ -> ());
    (try ignore (Unix.symlink (main_cache_dir ^ md5) cache1) with _ -> failwith "symlink failed");
    if Sys.file_exists (main_cache_dir ^ md5) then raise Exit;
    cache_oc := open_out (main_cache_dir ^ md5);
    prog "session" true;
    prog "parse" true;
    hh_svr 30 (tm, asm, gl)
  with Failure s -> Printf.fprintf !cache_oc "<br/>%s`%!" s;;

let hh_svr_line_exit ln = try hh_svr_line ln; flush !cache_oc; exit 0 with _ -> exit 0;;

(*
if Unix.fork () = 0 then Unix.execvp "dmtcp_command" [|""; "--checkpoint"|] else
try ignore (Unix.select [] [] [] 0.1) with _ -> ();;
*)
