needs "eval/tactics.ml";;
needs "eval/utils.ml";;
needs "eval/jobserver.ml";;
needs "HH/meson-tptp.ml";;

module Meson_eval = struct

(* the tactics to be evaluated are set by the Makefile *)
let tactics = ref []

(* timeout in seconds *)
let timeout = ref 10

let probno = ref 0

let pids = ref []

let process_prob ths g tactic =
  let eval = try Sys.getenv "EVAL" with _ -> "HH" in
  let outdir = eval ^ "/meson-eval/" ^ string_of_int !timeout ^ "s/" ^ tactic in
  mkdir_p outdir 0o755;
  let oname = outdir ^ "/" ^ string_of_int !probno ^ ".p" in
  Format.printf "Opening file %s ...\n%!" oname;
  let oc = open_out oname in
  Meson_thf.fprint_problem ths g oc;
  let run () =
    ignore (Tactics.run_tactic (Tactics.tactic_of_string tactic) ths g);
    Printf.fprintf oc "%% SZS Status Theorem\n%!" in
  log_execution oc !timeout run;
  Format.printf "Closing file %s.\n%!" oname;
  close_out oc

let hook ths g =
  incr probno;
  Format.printf "Processing problem %d...\n%!" !probno;
  try
    let fds = Jobserver.get_fds () in
    let add_job tac =
      pids := Jobserver.add_job fds (fun () -> process_prob ths g tac) !pids in
    List.iter add_job !tactics
  with Not_found -> begin
    Format.printf "Make jobserver not found!\n%!";
    List.iter (process_prob ths g) !tactics
  end

end
