needs "cop/utils.ml";;
needs "cop/strom.ml";;
needs "cop/term.ml";;
needs "cop/print.ml";;
needs "cop/substitution.ml";;
needs "cop/ncmatrix.ml";;
needs "cop/ncprint.ml";;
needs "cop/ncproof.ml";;
needs "cop/extclause.ml";;
needs "cop/ncdatabase.ml";;
needs "cop/cop.ml";;
needs "cop/ncsearch.ml";;
needs "cop/mapping.ml";;
needs "cop/ncrecon.ml";;
needs "cop/equality.ml";;
needs "cop/prepare.ml";;

module Nanocop = struct

open Cop
open Ncmatrix
open Utils

let start mat opt lim =
  if !verbose then Format.printf "Start %d\n%!" (lim+1);
  let mati = copy_matrix 0 mat in
  let sub0 = Subst.empty 1000000 in
  let hashek = Mapping.fol_of_literal [] [] Hashek.hashek_tm in
  Strom.of_list mati |> Strom.filter_map (function
      (((i, _), v) as idx, ([Lit lit] as cla)) when lit = hashek ->
        Ncsearch.prove_clause opt (sub0, 0) (mati, [], [i, 0], [], lim) cla
        |> Strom.next_opt >>= (fun (sub, prfs) -> Some (sub, Ncproof.Decomposition (cla, idx, prfs))
        )
    | _ -> None
    )
  |> Strom.next_opt >>= (fun (sub, prf) -> Some (fst sub, (Mat ((-1, 0), mati), prf)))

let fol_of_thm th =
  let lconsts = freesl (hyp th) in
  let tm = concl th in
  assert (List.for_all (fun x -> List.mem x lconsts) (frees tm));
  Mapping.fol_of_form [] lconsts tm



let NANOCOP_DEEPEN opt ths =
  if !verbose then begin
    Format.printf "NANOCOP_DEEPEN with ths:\n%!";
    List.iter (Format.printf "  th: %s\n%!" o string_of_thm) ths
  end;
  Mapping.reset_vars(); Mapping.reset_consts();
  let gen_eq = map GEN_ALL o Equality.create_eq_axs opt.add_eq_symmetry o map concl in
  let ths' =
    if List.exists (fun th -> concl th = Hashek.hashek_tm) ths
    then ths else Hashek.hashek_thm :: ths in
  let ths' = setify (ths' @ gen_eq ths') in
  let th = end_itlist CONJ ths' in
  let fol = fol_of_thm th in
  let mat = Ncmatrix.index_matrix (Ncmatrix.matrix_of_form fol) in
  if !verbose then (
    Format.printf "Matrix:\n%!";
    Ncprint.print_matrix_i mat);
  Ncdatabase.matrix2db mat;
  if !verbose then Format.printf "Matrix loaded.\n%!";
  let (sub, prfs) = Cop.strategy opt (start mat) in
  if !verbose then begin
    Format.printf "Theorem\n%!";
    Ncproof.print_proof_def sub prfs
  end;
  let data = Ncrecon.mk_recon_data (th, mat, sub) in
  let bot  = Ncrecon.bot_of_proof data [] th prfs in
  if !verbose then (Format.printf "Reconstruction:\n%!"; print_thm bot);
  assert (set_eq (hyp bot) (hyp th));
  assert (concl bot = `F`);
  bot


(* Debugging tactic. *)
let PRINT_TAC g = if !verbose then print_goal g; ALL_TAC g
let PRINT_ID_TAC s g = if !verbose then print_endline s; PRINT_TAC g

let PREPARE_NANOCOP_TAC ths =
    Prepare.PREPARE_COP_TAC ths THEN
    RULE_ASSUM_TAC(CONV_RULE(DEPTH_CONV(CHANGED_CONV(ASSOC_CONV DISJ_ASSOC)))) THEN
    RULE_ASSUM_TAC(CONV_RULE(DEPTH_CONV(CHANGED_CONV(ASSOC_CONV CONJ_ASSOC)))) THEN
    REPEAT (FIRST_X_ASSUM SUBST_VAR_TAC)

let PURE_NANOCOP opt gl =
  (FIRST_ASSUM CONTR_TAC ORELSE W(ACCEPT_TAC o NANOCOP_DEEPEN opt o map snd o fst)) gl

let GEN_NANOCOP_TAC opt ths = PREPARE_NANOCOP_TAC ths THEN PURE_NANOCOP opt

let NANOCOP_TAC ths = GEN_NANOCOP_TAC Cop.search_opts ths
let NANOCOP ths tm = prove(tm, NANOCOP_TAC ths)

end
