#load "unix.cma";;
#use "cop/nanocop.ml";;

let time f x =
  if not (!report_timing) then f x else
  let start_time = Unix.gettimeofday () in
  try let result = f x in
      let finish_time = Unix.gettimeofday () in
      report("CPU time (user): "^(string_of_float(finish_time -. start_time)));
      result
  with e ->
      let finish_time = Unix.gettimeofday () in
      Format.print_string("Failed after (user) CPU time of "^
                          (string_of_float(finish_time -. start_time))^": ");
      raise e;;


open Nanocop;;

verbose := false;;

let NCCOMP_TAC ths =
  GEN_NANOCOP_TAC Cop.({search_opts with cut1 = false; cut2 = false; cut3 = false}) ths;;

let NCCOMP ths tm = prove(tm, NCCOMP_TAC ths);;

let lemma_test = `~(!x. q x /\ !x. (~q x \/ p x) /\ !x. (~p x \/ ~q x))`;;
time (NCCOMP []) lemma_test;;

let ijcar16 = `p(a:t) /\ ((q(f(f(c:t))) /\ (!A. (~q(f(A)) \/ q(A:t))) /\ ~q(c)) \/ !A. (~p(A) \/ p(g(A)))) /\ !A. ~p(g(g(A)))`;;
time (NCCOMP []) (mk_neg ijcar16);;

let tab11 = `q (f b) /\ ~p a /\ !A:t. (p A /\ ~q (f A) \/ q A /\ ~r A /\ (~q a \/ r b))`;;
time (NCCOMP []) (mk_neg tab11);;

time (NCCOMP [APPEND]) `!h t. APPEND [(h : A)] t = CONS h t`;;
time (NCCOMP [LE; LT_NZ; NOT_LE; ONE; LT]) (concl LT);;

prioritize_real ();;
time (NCCOMP [real_lt; REAL_LE_TOTAL]) `!x y. x < y ==> x <= y`;;

time (NANOCOP [REAL_LE_LNEG; REAL_ADD_AC; real_lt]) (concl REAL_LT_RNEG);;
