module Print = struct

let rec pp_interleave sep fn f = function
    x :: (_ :: _ as xs) -> fn f x; sep f; pp_interleave sep fn f xs
  | x :: [] -> fn f x
  | [] -> ()

let pp_iter sep = pp_interleave (fun f -> pp_print_string f sep)

let pp_iter_sp sep fn =
  let sep f = pp_print_char f ' '; pp_print_string f sep; pp_print_space f () in
  pp_interleave sep fn

let pp_enclose start stop fn f x =
  pp_print_string f start; fn f x; pp_print_string f stop

let pp_enclose_cut start stop fn =
  pp_enclose start stop (fun f x -> fn f x; pp_print_cut f ())

let pp_enum sep start stop fn =
  pp_enclose start stop (pp_iter sep fn)

let pp_with_box i fn f x =
  pp_open_box f i; fn f x; pp_close_box f ()

let pp_nl fn f x = fn f x; pp_print_newline f ()

let pp_print_null f x = ()


let pp_print_var f i =
  pp_print_char f (Char.chr (65 + i mod 26));
  if i > 25 then pp_print_int f (i / 26)

let rec pp_print_fterm f = function
    Fvar i -> pp_print_var f i
  | Fnapp (i, l) ->
      pp_print_term f (Meson.hol_of_const i);
      if l <> [] then pp_enum "," "(" ")" pp_print_fterm f l

let pp_print_lit f (i, l) =
  if i < 0 then pp_print_char f '~';
  pp_print_term f (Meson.hol_of_const (abs i));
  if l <> [] then pp_enum "," "(" ")" pp_print_fterm f l

let string_of_fterm = print_to_string pp_print_fterm
let string_of_lit = print_to_string pp_print_lit
let string_of_lits = print_to_string (pp_iter ", " pp_print_lit)

end
