module Search = struct

open Print
open Fol_term
open Strom
open Proof
open Utils
open Cop

let cut p xs ys = Strom.(if p then cut xs ys else append xs ys)

let rec prove_lit opt sub (path, lem, lim) lit =
  (*if !verbosenc then (Format.printf "%d %s\n%!" !plits (string_of_lit (pred_of_lit lit)); incr plits);*)

  if !verbose then Format.printf "Lit: %s\n%!" (string_of_lit (Subst.inst_lit (fst sub) lit));
  if !verbose then Format.printf "Path: %s\n%!" (string_of_lits (List.map (Subst.inst_lit (fst sub)) path));
  if !verbose then Format.printf "Lemmas: %s\n%!" (string_of_lits (List.map (Subst.inst_lit (fst sub)) lem));

  let neglit = negate_lit lit in
  let lemmas =
    if opt.improved_lemmas then begin
      if List.exists (Subst.eq sub lit) lem then
       (if !verbose then Format.printf "lemma\n%!"; cons (sub, lem, Lemma) nil) else nil
    end
    else
      Strom.of_list lem |> Strom.filter (Subst.eq sub lit) |>
      Strom.map (fun l -> if !verbose then Format.printf "lemma\n%!"; (sub, lit :: lem, Lemma))
  and reductions = Strom.filter_map
    (fun p ->
      if !verbose then Format.printf "Reduction try %s (for lit %s, lim %d)\n%!" (string_of_lit p) (string_of_lit (Subst.inst_lit (fst sub) lit)) lim;
      Subst.unify sub neglit p >>=
      (fun sub1 -> if !verbose then Format.printf "Reduction works\n%!";
        Some (sub1, lit :: lem, Reduction))
    ) (Strom.of_list path)
  and extensions = Database.db_entries neglit
    |> Strom.of_list
    |> Strom.concat_map (fun ((_, _, vars, hsh) as contra) ->
(*
      if !verbose then Format.printf "Extension try %s (for lit %s, lim %d)\n%!" (Hashtbl.find Database.Clausal.no_contr hsh) (string_of_lit (Subst.inst_lit (fst sub) lit)) lim;
*)
      if lim < 0 && vars > 0 then nil
      else match Subst.unify_rename sub (snd lit) contra with
        Some (sub1, cla1) ->
(*
          if !verbose then Format.printf "Extension clause %s found\n%!" (List.map (Subst.inst_lit (fst sub1)) cla1 |> string_of_clause);
          incr Stats.infer;
*)

          prove_clause opt sub1 (lit :: path, lem, lim - 1) (cla1, hsh)
          |> Strom.map (fun (sub2, prfs) -> (sub2, lit :: lem, Extension (hsh, prfs)))
      | None -> nil) in
  cut opt.cut1 lemmas (cut opt.cut2 reductions (cut opt.cut3 extensions nil))
and prove_clause opt sub (path, lem, lim) (cl, cl_hsh) = match cl with
    lit :: lits ->
    if (List.exists (fun x -> List.exists (Subst.eq sub x) path)) cl then (if !verbose then Format.printf "regularity\n%!"; nil)
    else
    prove_lit opt sub (path, lem, lim) lit
    (*|> Litdata.trace_proofs lit cl_hsh*)
    |> Strom.concat_map
       (fun (sub1, lem1, prf1) -> prove_clause opt sub1 (path, lem1, lim) (lits, cl_hsh) |> Strom.map
       (fun (sub2, prfs) -> (sub2, (lit, prf1) :: prfs)))
  | [] -> cons (sub, []) nil

end
