let parallelise f x n_processes =
  (* PIDs of running threads *)
  let thrs = ref [] in
  let state = ref x in

  let add_thread () = match f (!state) with
    | None -> ()
    | Some (pid, state') ->
        assert (not (List.mem pid !thrs));
        (*Format.printf "New process with pid = %d\n%!" pid;*)
        thrs := pid :: !thrs;
        state := state' in

  (* create initial thread pool, and whenever a thread is done, start new one *)
  for i = 0 to n_processes-1 do add_thread () done;
  while !thrs <> [] do
    let pid = fst (Unix.waitpid [] (-1)) in
    (* remove finished thread from thread list *)
    thrs := List.filter ((<>) pid) !thrs;
    add_thread ();
  done;
  !state
;;

(* if some list item is left, run function in new thread and return rest list *)
let fork_list (f : 'a -> unit) = function
    [] -> None
  | x :: xs ->
      (match Unix.fork () with
        0 -> f x; exit 0
      | pid -> Some (pid, xs))
;;

let par_iter f xs n = ignore (parallelise (fork_list f) xs n)


let add_job thrs max (f : unit -> unit) =
  let add_thread f =
     match Unix.fork () with
       0 -> f (); exit 0
     | pid ->
         assert (not (List.mem pid !thrs));
         (*Format.printf "New process with pid = %d\n%!" pid;*)
         thrs := pid :: !thrs in

  assert (List.length !thrs <= max);
  if List.length !thrs = max then begin
    let pid = fst (Unix.waitpid [] (-1)) in
    (* remove finished thread from thread list *)
    assert (List.mem pid !thrs);
    thrs := List.filter ((<>) pid) !thrs
  end;
  add_thread f
