needs "cop/nanocop.ml";;
needs "cop/leancop.ml";;
needs "metis.ml";;
needs "metis2/metis_tac.ml";;


module Tactics =
struct

type splitlim_t = Splitlimit of int;;
type cut_t = Cut of bool;;

type tactic_t =
    Meson of splitlim_t
  | Metis of splitlim_t
  | Metis2 of splitlim_t
  | Leancop of cut_t
  | Nanocop of cut_t
;;

let tactic_of_string = function
    "meson-sl0"  -> Meson  (Splitlimit 0)
  | "meson-sl8"  -> Meson  (Splitlimit 8)
  | "metis-sl0"  -> Metis  (Splitlimit 0)
  | "metis-sl8"  -> Metis  (Splitlimit 8)
  | "metis2-sl0" -> Metis2 (Splitlimit 0)
  | "metis2-sl8" -> Metis2 (Splitlimit 8)
  | "leancop-cut"   -> Leancop (Cut true)
  | "leancop-nocut" -> Leancop (Cut false)
  | "nanocop-cut"   -> Nanocop (Cut true)
  | "nanocop-nocut" -> Nanocop (Cut false)
  | p -> failwith ("unknown tactic " ^ p)
;;

let run_tactic = function
    Meson (Splitlimit sl) ->
      meson_split_limit := sl;
      (* increase the maximum depth from 50 (default) to 1000 *)
      let ASM_MESON_TAC = GEN_MESON_TAC 0 1000 1 in
      ASM_MESON_TAC
  | Metis (Splitlimit sl) -> Metis.split_limit := sl; ASM_METIS_TAC
  | Metis2 (Splitlimit sl) -> Metis2.split_limit := sl; Metis2.ASM_METIS_TAC
  | Leancop (Cut c) ->
      verbose := false;
      Leancop.GEN_LEANCOP_TAC Cop.({search_opts with cut1 = c; cut2 = c; cut3 = c})
  | Nanocop (Cut c) ->
      verbose := false;
      Nanocop.GEN_NANOCOP_TAC Cop.({search_opts with cut1 = c; cut2 = c; cut3 = c})
;;

end
