(* Running OCaml functions with a fixed timeout. *)
(* Copyright (C) 2017 by Michael Färber under
   Version 3 of the GNU GPL or any later version. *)

#load "unix.cma";;

type ('a, 'b) either = Left of 'a | Right of 'b

unset_jrh_lexer;;

let with_timeout timeout (f : unit -> 'a) =
  let pipe_r, pipe_w = Unix.pipe () in
  match Unix.fork () with
    0 ->
      Unix.close pipe_r;
      ignore (Unix.alarm timeout);
      let y = (try Right (f ()) with e -> Left e) in
      ignore (Unix.alarm 0);
      let oc = Unix.out_channel_of_descr pipe_w in
      Marshal.to_channel oc y [];
      close_out oc;
      exit 0
  | pid ->
      Unix.close pipe_w;
      (*Format.printf "Forked process %d\n%!" pid;*)
      let pid', status = Unix.waitpid [] pid in
      assert (pid = pid');
      begin match status with
        Unix.WEXITED 0 ->
          let ic = Unix.in_channel_of_descr pipe_r in
          let result = (Marshal.from_channel ic : (exn, 'a) either) in
          close_in ic;
          (match result with Left e -> raise e | Right y -> y)
      | Unix.WEXITED n ->
          failwith ("Process " ^ string_of_int pid ^ " exited with return code: " ^ string_of_int n)
      | Unix.WSIGNALED s | Unix.WSTOPPED s ->
          if s = Sys.sigalrm then failwith "timeout"
          else failwith ("Process " ^ string_of_int pid ^ " exited with signal: " ^ string_of_int s)
      end
;;

set_jrh_lexer;;
